import importlib.util
import json
from pathlib import Path
from typing import Any

import numpy as np

# Tokenizers imports moved to be optional


class BaseTokenizer:
    """Abstract Tokenizer class"""

    @property
    def vocab_size(self) -> int:
        """Returns the vocabulary size of the tokenizer

        :return: Vocabulary size of tokenizer
        :rtype: int
        """
        raise NotImplementedError

    @property
    def config(self) -> dict:
        """The tokenizer config

        :return: dictionary of tokenizer config
        :rtype: dict
        """
        raise NotImplementedError

    def save(self, path: str | Path) -> None:
        """Saves the configuration of the tokenizer

        Saves these files:
        - tokenizer.json: saved configuration of the tokenizer

        :param path: Path to save the tokenizer to
        :type path: str, :class:`pathlib.Path`
        """
        if isinstance(path, str):
            path = Path(path)

        with open(path, "w") as fp:
            json.dump(self.config, fp)

    @classmethod
    def load(cls, path: str | Path) -> "BaseTokenizer":
        """Returns a :class:`bm25_engine.tokenizer.BaseTokenizer` object from saved configuration

        Requires these files:
        - tokenizer.json: saved configuration of the tokenizer

        :param path: Path to load the tokenizer from
        :type path: str, :class:`pathlib.Path`
        :returns: Configured BaseTokenizer
        :rtype: BaseTokenizer
        """
        if isinstance(path, str):
            path = Path(path)

        with open(path) as fp:
            config = json.load(fp)
        return cls(**config)

    def tokenize(self, texts: str | list[str], pad: bool = True) -> np.ndarray:
        raise NotImplementedError


class HashTokenizer(BaseTokenizer):
    def __init__(self) -> None:
        super().__init__()


class PretrainedTokenizer(BaseTokenizer):
    """Wrapper for HuggingFace tokenizers, representing a pretrained tokenizer (i.e. bert-base-uncased).
    Extends the :class:`semantic_router.tokenizers.BaseTokenizer` class.

    :param tokenizer: Binding for HuggingFace Rust tokenizers
    :type tokenizer: class:`tokenizers.Tokenizer`
    :param add_special_tokens: Whether to accept special tokens from the tokenizer (i.e. `[PAD]`)
    :type add_special_tokens: bool
    :param pad: Whether to pad the input to a consistent length (using `[PAD]` tokens)
    :type pad: bool
    :param model_ident: HuggingFace ID of the model (i.e. `bert-base-uncased`)
    :type model_ident: str
    """

    add_special_tokens: bool
    pad: bool
    model_ident: str

    def __init__(
        self,
        model_ident: str,
        custom_normalizer: Any = None,
        add_special_tokens: bool = False,
        pad: bool = True,
    ) -> None:
        """Constructor method"""
        # Check if tokenizers is available
        if importlib.util.find_spec("tokenizers") is None:
            raise ImportError(
                "The 'tokenizers' package is required for PretrainedTokenizer but not installed. "
                "Please install it with `pip install tokenizers`."
            )

        # Import tokenizers only when needed
        from tokenizers import Tokenizer

        super().__init__()
        self.add_special_tokens = add_special_tokens
        self.model_ident = model_ident
        self.tokenizer = Tokenizer.from_pretrained(model_ident)
        self.pad = pad
        if custom_normalizer:
            self.tokenizer.normalizer = custom_normalizer  # type: ignore
        if pad:
            self.tokenizer.enable_padding(direction="right", pad_id=0)

    @property
    def vocab_size(self):
        """Returns the vocabulary size of the tokenizer

        :return: Vocabulary size of tokenizer
        :rtype: int
        """
        return self.tokenizer.get_vocab_size()

    @property
    def config(self) -> dict:
        """The tokenizer config

        :return: dictionary of tokenizer config
        :rtype: dict
        """
        return {
            "model_ident": self.model_ident,
            "add_special_tokens": self.add_special_tokens,
            "pad": self.pad,
        }

    def tokenize(self, texts: str | list[str], pad: bool = True) -> np.ndarray:
        """Tokenizes a string or list of strings into a 2D :class:`numpy.ndarray` of token ids

        :param texts: Texts to be tokenized
        :type texts: str, list
        :param pad: unused here (configured in the constructor)
        :type pad: bool
        :return: 2D numpy array representing token ids
        :rtype: class:`numpy.ndarray`
        """
        if isinstance(texts, str):
            texts = [texts]

        encodings = self.tokenizer.encode_batch_fast(
            texts, add_special_tokens=self.add_special_tokens
        )

        return np.array([e.ids for e in encodings])


class TokenizerFactory:
    """Tokenizer factory class"""

    @staticmethod
    def get(type_: str, **tokenizer_kwargs) -> BaseTokenizer:
        r"""Get a configured :class:`bm25_engine.tokenizer.BaseTokenizer`

        :param type_: Tokenizer type to instantiate
        :type type_: str
        :param \**kwargs: kwargs to be passed to Tokenizer constructor
        :return: Tokenizer
        :rtype: `bm25_engine.tokenizer.BaseTokenizer`
        """
        match type_:
            case "pretrained":
                return PretrainedTokenizer(**tokenizer_kwargs)
            case _:
                return PretrainedTokenizer(**tokenizer_kwargs)



"""
This module provides the BedrockEncoder class for generating embeddings using Amazon's Bedrock Platform.

The BedrockEncoder class is a subclass of DenseEncoder and utilizes the TextEmbeddingModel from the
Amazon's Bedrock Platform to generate embeddings for given documents. It requires an AWS Access Key ID
and AWS Secret Access Key and supports customization of the pre-trained model, score threshold, and region.

Example usage:

    from semantic_router.encoders.bedrock_encoder import BedrockEncoder

    encoder = BedrockEncoder(access_key_id="your-access-key-id", secret_access_key="your-secret-key", region="your-region")
    embeddings = encoder(["document1", "document2"])

Classes:
    BedrockEncoder: A class for generating embeddings using the Bedrock Platform.
"""

import json
import os
from time import sleep
from typing import Any, Dict, List, Optional, Union

import tiktoken

from semantic_router.encoders import DenseEncoder
from semantic_router.utils.defaults import EncoderDefault
from semantic_router.utils.logger import logger


class BedrockEncoder(DenseEncoder):
    """Dense encoder using Amazon Bedrock embedding API. Requires an AWS Access Key ID
    and AWS Secret Access Key.

    The BedrockEncoder class is a subclass of DenseEncoder and utilizes the
    TextEmbeddingModel from the Amazon's Bedrock Platform to generate embeddings for
    given documents. It supports customization of the pre-trained model, score
    threshold, and region.

    Example usage:

    ```python
    from semantic_router.encoders.bedrock_encoder import BedrockEncoder

    encoder = BedrockEncoder(
        access_key_id="your-access-key-id",
        secret_access_key="your-secret-key",
        region="your-region"
    )
    embeddings = encoder(["document1", "document2"])
    ```
    """

    client: Any = None
    type: str = "bedrock"
    input_type: Optional[str] = "search_query"
    name: str
    access_key_id: Optional[str] = None
    secret_access_key: Optional[str] = None
    session_token: Optional[str] = None
    region: Optional[str] = None

    def __init__(
        self,
        name: str = EncoderDefault.BEDROCK.value["embedding_model"],
        input_type: Optional[str] = "search_query",
        score_threshold: float = 0.3,
        access_key_id: Optional[str] = None,
        secret_access_key: Optional[str] = None,
        session_token: Optional[str] = None,
        region: Optional[str] = None,
    ):
        """Initializes the BedrockEncoder.

        :param name: The name of the pre-trained model to use for embedding.
            If not provided, the default model specified in EncoderDefault will
            be used.
        :type name: str
        :param input_type: The type of input to use for the embedding.
            If not provided, the default input type specified in EncoderDefault will
            be used.
        :type input_type: str
        :param score_threshold: The threshold for similarity scores.
        :type score_threshold: float
        :param access_key_id: The AWS access key id for an IAM principle.
            If not provided, it will be retrieved from the access_key_id
            environment variable.
        :type access_key_id: str
        :param secret_access_key: The secret access key for an IAM principle.
            If not provided, it will be retrieved from the AWS_SECRET_KEY
            environment variable.
        :type secret_access_key: str
        :param session_token: The session token for an IAM principle.
            If not provided, it will be retrieved from the AWS_SESSION_TOKEN
            environment variable.
        :param region: The location of the Bedrock resources.
            If not provided, it will be retrieved from the AWS_REGION
            environment variable, defaulting to "us-west-1"
        :type region: str
        :raises ValueError: If the Bedrock Platform client fails to initialize.
        """
        super().__init__(name=name, score_threshold=score_threshold)
        self.access_key_id = self.get_env_variable("AWS_ACCESS_KEY_ID", access_key_id)
        self.secret_access_key = self.get_env_variable(
            "AWS_SECRET_ACCESS_KEY", secret_access_key
        )
        self.session_token = self.get_env_variable("AWS_SESSION_TOKEN", session_token)
        self.region = self.get_env_variable(
            "AWS_DEFAULT_REGION", region, default="us-west-1"
        )
        self.input_type = input_type
        try:
            self.client = self._initialize_client(
                self.access_key_id,
                self.secret_access_key,
                self.session_token,
                self.region,
            )
        except Exception as e:
            raise ValueError(f"Bedrock client failed to initialise. Error: {e}") from e

    def _initialize_client(
        self, access_key_id, secret_access_key, session_token, region
    ):
        """Initializes the Bedrock client.

        :param access_key_id: The Amazon access key ID.
        :type access_key_id: str
        :param secret_access_key: The Amazon secret key.
        :type secret_access_key: str
        :param region: The location of the AI Platform resources.
        :type region: str
        :returns: An instance of the TextEmbeddingModel client.
        :rtype: Any
        :raises ImportError: If the required Bedrock libraries are not
            installed.
            ValueError: If the Bedrock client fails to initialize.
        """
        try:
            import boto3
        except ImportError:
            raise ImportError(
                "Please install Amazon's Boto3 client library to use the BedrockEncoder. "
                "You can install them with: "
                "`pip install boto3`"
            )
        access_key_id = access_key_id or os.getenv("AWS_ACCESS_KEY_ID")
        aws_secret_key = secret_access_key or os.getenv("AWS_SECRET_ACCESS_KEY")
        region = region or os.getenv("AWS_DEFAULT_REGION", "us-west-2")
        if access_key_id is None:
            raise ValueError("AWS access key ID cannot be 'None'.")
        if aws_secret_key is None:
            raise ValueError("AWS secret access key cannot be 'None'.")
        session = boto3.Session(
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            aws_session_token=session_token,
        )
        try:
            bedrock_client = session.client(
                "bedrock-runtime",
                region_name=region,
            )
        except Exception as err:
            raise ValueError(
                f"The Bedrock client failed to initialize. Error: {err}"
            ) from err
        return bedrock_client

    def __call__(
        self, docs: List[Union[str, Dict]], model_kwargs: Optional[Dict] = None
    ) -> List[List[float]]:
        """Generates embeddings for the given documents.

        :param docs: A list of strings representing the documents to embed.
        :type docs: list[str]
        :param model_kwargs: A dictionary of model-specific inference parameters.
        :type model_kwargs: dict
        :returns: A list of lists, where each inner list contains the embedding values for a
            document.
        :rtype: list[list[float]]
        :raises ValueError: If the Bedrock Platform client is not initialized or if the
            API call fails.
        """
        try:
            from botocore.exceptions import ClientError
        except ImportError:
            raise ImportError(
                "Please install Amazon's Botocore client library to use the BedrockEncoder. "
                "You can install them with: "
                "`pip install botocore`"
            )
        if self.client is None:
            raise ValueError("Bedrock client is not initialised.")
        max_attempts = 3
        for attempt in range(max_attempts):
            try:
                embeddings = []
                if self.name and "amazon" in self.name:
                    for doc in docs:
                        embedding_body = {}

                        if isinstance(doc, dict):
                            embedding_body["inputText"] = doc.get("text")
                            embedding_body["inputImage"] = doc.get(
                                "image"
                            )  # expects a base64-encoded image
                        else:
                            embedding_body["inputText"] = doc

                        # Add model-specific inference parameters
                        if model_kwargs:
                            embedding_body = embedding_body | model_kwargs

                        # Clean up null values
                        embedding_body = {k: v for k, v in embedding_body.items() if v}

                        # Format payload
                        embedding_body_payload: str = json.dumps(embedding_body)

                        response = self.client.invoke_model(
                            body=embedding_body_payload,
                            modelId=self.name,
                            accept="application/json",
                            contentType="application/json",
                        )
                        response_body = json.loads(response.get("body").read())
                        embeddings.append(response_body.get("embedding"))
                elif self.name and "cohere" in self.name:
                    chunked_docs = self.chunk_strings(docs)
                    for chunk in chunked_docs:
                        chunk = {"texts": chunk, "input_type": self.input_type}

                        # Add model-specific inference parameters
                        # Note: if specified, input_type will be overwritten by model_kwargs
                        if model_kwargs:
                            chunk = chunk | model_kwargs

                        # Format payload
                        chunk = json.dumps(chunk)

                        response = self.client.invoke_model(
                            body=chunk,
                            modelId=self.name,
                            accept="*/*",
                            contentType="application/json",
                        )
                        response_body = json.loads(response.get("body").read())
                        chunk_embeddings = response_body.get("embeddings")
                        embeddings.extend(chunk_embeddings)
                else:
                    raise ValueError("Unknown model name")
                return embeddings
            except ClientError as error:
                if attempt < max_attempts - 1:
                    if error.response["Error"]["Code"] == "ExpiredTokenException":
                        logger.warning(
                            "Session token has expired. Retrying initialisation."
                        )
                        try:
                            self.session_token = os.getenv("AWS_SESSION_TOKEN")
                            self.client = self._initialize_client(
                                self.access_key_id,
                                self.secret_access_key,
                                self.session_token,
                                self.region,
                            )
                        except Exception as e:
                            raise ValueError(
                                f"Bedrock client failed to reinitialise. Error: {e}"
                            ) from e
                    sleep(2**attempt)
                    logger.warning(f"Retrying in {2**attempt} seconds...")
                raise ValueError(
                    f"Retries exhausted, Bedrock call failed. Error: {error}"
                ) from error
            except Exception as e:
                raise ValueError(f"Bedrock call failed. Error: {e}") from e
        raise ValueError("Bedrock call failed to return embeddings.")

    def chunk_strings(self, strings, MAX_WORDS=20):
        """Breaks up a list of strings into smaller chunks.

        :param strings: A list of strings to be chunked.
        :type strings: list
        :param max_chunk_size: The maximum size of each chunk. Default is 20.
        :type max_chunk_size: int
        :returns: A list of lists, where each inner list contains a chunk of strings.
        :rtype: list[list[str]]
        """
        encoding = tiktoken.get_encoding("cl100k_base")
        chunked_strings = []
        for text in strings:
            encoded_text = encoding.encode(text)
            chunks = [
                encoding.decode(encoded_text[i : i + MAX_WORDS])
                for i in range(0, len(encoded_text), MAX_WORDS)
            ]
            chunked_strings.append(chunks)
        return chunked_strings

    @staticmethod
    def get_env_variable(var_name, provided_value, default=None):
        """Retrieves environment variable or uses a provided value.

        :param var_name: The name of the environment variable.
        :type var_name: str
        :param provided_value: The provided value to use if not None.
        :type provided_value: Optional[str]
        :param default: The default value if the environment variable is not set.
        :type default: Optional[str]
        :returns: The value of the environment variable or the provided/default value.
        :rtype: str
        :raises ValueError: If no value is provided and the environment variable is not set.
        """
        if provided_value is not None:
            return provided_value
        value = os.getenv(var_name, default)
        if value is None:
            if var_name == "AWS_SESSION_TOKEN":
                return None
            raise ValueError(f"No {var_name} provided")
        return value



import asyncio
import string
from collections import Counter
from typing import Dict, List

import numpy as np

from semantic_router.encoders import SparseEncoder
from semantic_router.encoders.base import FittableMixin
from semantic_router.route import Route
from semantic_router.schema import SparseEmbedding


class TfidfEncoder(SparseEncoder, FittableMixin):
    idf: np.ndarray = np.array([])
    # TODO: add option to use default params like with BM25Encoder
    word_index: Dict = {}

    def __init__(self, name: str | None = None):
        if name is None:
            name = "tfidf"
        super().__init__(name=name)
        self.word_index = {}
        self.idf = np.array([])

    def __call__(self, docs: List[str]) -> list[SparseEmbedding]:
        if len(self.word_index) == 0 or self.idf.size == 0:
            raise ValueError("Vectorizer is not initialized.")
        if len(docs) == 0:
            raise ValueError("No documents to encode.")

        docs = [self._preprocess(doc) for doc in docs]
        tf = self._compute_tf(docs)
        tfidf = tf * self.idf
        return self._array_to_sparse_embeddings(tfidf)

    async def acall(self, docs: List[str]) -> List[SparseEmbedding]:
        return await asyncio.to_thread(lambda: self.__call__(docs))

    def fit(self, routes: List[Route]):
        """Trains the encoder weights on the provided routes.

        :param routes: List of routes to train the encoder on.
        :type routes: List[Route]
        """
        self._fit_validate(routes=routes)
        docs = []
        for route in routes:
            for doc in route.utterances:
                docs.append(self._preprocess(doc))  # type: ignore
        self.word_index = self._build_word_index(docs)
        if len(self.word_index) == 0:
            raise ValueError(f"Too little data to fit {self.__class__.__name__}.")
        self.idf = self._compute_idf(docs)

    def _fit_validate(self, routes: List[Route]):
        if not isinstance(routes, list) or not isinstance(routes[0], Route):
            raise TypeError("`routes` parameter must be a list of Route objects.")

    def _build_word_index(self, docs: List[str]) -> Dict:
        words = set()
        for doc in docs:
            for word in doc.split():
                words.add(word)
        word_index = {word: i for i, word in enumerate(words)}
        return word_index

    def _compute_tf(self, docs: List[str]) -> np.ndarray:
        if len(self.word_index) == 0:
            raise ValueError("Word index is not initialized.")
        tf = np.zeros((len(docs), len(self.word_index)))
        for i, doc in enumerate(docs):
            word_counts = Counter(doc.split())
            for word, count in word_counts.items():
                if word in self.word_index:
                    tf[i, self.word_index[word]] = count
        # L2 normalization
        tf = tf / np.linalg.norm(tf, axis=1, keepdims=True)
        return tf

    def _compute_idf(self, docs: List[str]) -> np.ndarray:
        if len(self.word_index) == 0:
            raise ValueError("Word index is not initialized.")
        idf = np.zeros(len(self.word_index))
        for doc in docs:
            words = set(doc.split())
            for word in words:
                if word in self.word_index:
                    idf[self.word_index[word]] += 1
        idf = np.log(len(docs) / (idf + 1))
        return idf

    def _preprocess(self, doc: str) -> str:
        lowercased_doc = doc.lower()
        no_punctuation_doc = lowercased_doc.translate(
            str.maketrans("", "", string.punctuation)
        )
        return no_punctuation_doc



import os
from asyncio import sleep as asleep
from time import sleep
from typing import Any, Callable, Dict, List, Optional, Union

import httpx
import openai
from openai import OpenAIError
from openai._types import NotGiven
from openai.types import CreateEmbeddingResponse

from semantic_router.encoders import DenseEncoder
from semantic_router.utils.defaults import EncoderDefault
from semantic_router.utils.logger import logger


class AzureOpenAIEncoder(DenseEncoder):
    """Encoder for Azure OpenAI API.

    This class provides functionality to encode text documents using the Azure OpenAI API.
    It supports customization of the score threshold for filtering or processing the embeddings.
    """

    client: Optional[openai.AzureOpenAI] = None
    async_client: Optional[openai.AsyncAzureOpenAI] = None
    dimensions: Union[int, NotGiven] = NotGiven()
    type: str = "azure"
    deployment_name: str | None = None
    max_retries: int = 3

    def __init__(
        self,
        name: Optional[str] = None,
        azure_endpoint: str | None = None,
        api_version: str | None = None,
        api_key: str | None = None,
        azure_ad_token: str | None = None,
        azure_ad_token_provider: Callable[[], str] | None = None,
        http_client_options: Optional[Dict[str, Any]] = None,
        deployment_name: str = EncoderDefault.AZURE.value["deployment_name"],
        score_threshold: float = 0.82,
        dimensions: Union[int, NotGiven] = NotGiven(),
        max_retries: int = 3,
    ):
        """Initialize the AzureOpenAIEncoder.

        :param azure_endpoint: The endpoint for the Azure OpenAI API.
            Example: `"https://accountname.openai.azure.com"`
        :type azure_endpoint: str, optional

        :param api_version: The version of the API to use.
            Example: `"2025-02-01-preview"`
        :type api_version: str, optional

        :param api_key: The API key for the Azure OpenAI API.
        :type api_key: str, optional

        :param azure_ad_token: The Azure AD/Entra ID token for authentication.
            https://www.microsoft.com/en-us/security/business/identity-access/microsoft-entra-id
        :type azure_ad_token: str, optional

        :param azure_ad_token_provider: A callable function that returns an Azure AD/Entra ID token.
        :type azure_ad_token_provider: Callable[[], str], optional

        :param http_client_options: Dictionary of options to configure httpx client
            Example:
            ```
                {
                    "proxies": "http://proxy.server:8080",
                    "timeout": 20.0,
                    "headers": {"Authorization": "Bearer xyz"}
                }
            ```
        :type http_client_options: Dict[str, Any], optional

        :param deployment_name: The name of the model deployment to use.
        :type deployment_name: str, optional

        :param score_threshold: The score threshold for filtering embeddings.
            Default is `0.82`.
        :type score_threshold: float, optional

        :param dimensions: The number of dimensions for the embeddings. If not given, it defaults to the model's default setting.
        :type dimensions: int, optional

        :param max_retries: The maximum number of retries for API calls in case of failures.
            Default is `3`.
        :type max_retries: int, optional
        """
        if name is None:
            name = deployment_name
            if name is None:
                name = EncoderDefault.AZURE.value["embedding_model"]
        super().__init__(name=name, score_threshold=score_threshold)

        azure_endpoint = azure_endpoint or os.getenv("AZURE_OPENAI_ENDPOINT")
        if not azure_endpoint:
            raise ValueError("No Azure OpenAI endpoint provided.")

        api_version = api_version or os.getenv("AZURE_OPENAI_API_VERSION")
        if not api_version:
            raise ValueError("No Azure OpenAI API version provided.")

        if not (
            azure_ad_token
            or azure_ad_token_provider
            or api_key
            or os.getenv("AZURE_OPENAI_API_KEY")
        ):
            raise ValueError(
                "No authentication method provided. Please provide either `azure_ad_token`, "
                "`azure_ad_token_provider`, or `api_key`."
            )

        # Only check API Key if no AD token or provider is used
        if not azure_ad_token and not azure_ad_token_provider:
            api_key = api_key or os.getenv("AZURE_OPENAI_API_KEY")
            if not api_key:
                raise ValueError("No Azure OpenAI API key provided.")

        self.deployment_name = deployment_name

        # set dimensions to support openai embed 3 dimensions param
        self.dimensions = dimensions

        if max_retries is not None:
            self.max_retries = max_retries

        # Only create HTTP clients if options are provided
        sync_http_client = (
            httpx.Client(**http_client_options) if http_client_options else None
        )
        async_http_client = (
            httpx.AsyncClient(**http_client_options) if http_client_options else None
        )

        assert azure_endpoint is not None and self.deployment_name is not None

        try:
            self.client = openai.AzureOpenAI(
                azure_endpoint=azure_endpoint,
                api_version=api_version,
                api_key=api_key,
                azure_ad_token=azure_ad_token,
                azure_ad_token_provider=azure_ad_token_provider,
                http_client=sync_http_client,
            )
            self.async_client = openai.AsyncAzureOpenAI(
                azure_endpoint=azure_endpoint,
                api_version=api_version,
                api_key=api_key,
                azure_ad_token=azure_ad_token,
                azure_ad_token_provider=azure_ad_token_provider,
                http_client=async_http_client,
            )

        except Exception as e:
            logger.error("OpenAI API client failed to initialize. Error: %s", e)
            raise ValueError(
                f"OpenAI API client failed to initialize. Error: {e}"
            ) from e

    def __call__(self, docs: List[str]) -> List[List[float]]:
        """Encode a list of documents into embeddings using the Azure OpenAI API.

        :param docs: The documents to encode.
        :type docs: List[str]
        :return: The embeddings for the documents.
        :rtype: List[List[float]]
        """
        if self.client is None:
            raise ValueError("Azure OpenAI client is not initialized.")
        embeds = None

        # Exponential backoff
        for j in range(self.max_retries + 1):
            try:
                embeds = self.client.embeddings.create(
                    input=docs,
                    model=str(self.deployment_name),
                    dimensions=self.dimensions,
                )
                if embeds.data:
                    break
            except OpenAIError as e:
                logger.error("Exception occurred", exc_info=True)
                if self.max_retries != 0 and j < self.max_retries:
                    sleep(2**j)
                    logger.warning(
                        "Retrying in %d seconds due to OpenAIError: %s", 2**j, e
                    )
                else:
                    raise
            except Exception as e:
                logger.error("Azure OpenAI API call failed. Error: %s", e)
                raise ValueError(f"Azure OpenAI API call failed. Error: {e}") from e

        if (
            not embeds
            or not isinstance(embeds, CreateEmbeddingResponse)
            or not embeds.data
        ):
            raise ValueError("No embeddings returned.")

        embeddings = [embeds_obj.embedding for embeds_obj in embeds.data]
        return embeddings

    async def acall(self, docs: List[str]) -> List[List[float]]:
        """Encode a list of documents into embeddings using the Azure OpenAI API asynchronously.

        :param docs: The documents to encode.
        :type docs: List[str]
        :return: The embeddings for the documents.
        :rtype: List[List[float]]
        """
        if self.async_client is None:
            raise ValueError("Azure OpenAI async client is not initialized.")
        embeds = None
        # Exponential backoff
        for j in range(self.max_retries + 1):
            try:
                embeds = await self.async_client.embeddings.create(
                    input=docs,
                    model=str(self.deployment_name),
                    dimensions=self.dimensions,
                )
                if embeds.data:
                    break
            except OpenAIError as e:
                logger.error("Exception occurred", exc_info=True)
                if self.max_retries != 0 and j < self.max_retries:
                    await asleep(2**j)
                    logger.warning(
                        "Retrying in %d seconds due to OpenAIError: %s", 2**j, e
                    )
                else:
                    raise
            except Exception as e:
                logger.error("Azure OpenAI API call failed. Error: %s", e)
                raise ValueError(f"Azure OpenAI API call failed. Error: {e}") from e

        if (
            not embeds
            or not isinstance(embeds, CreateEmbeddingResponse)
            or not embeds.data
        ):
            raise ValueError("No embeddings returned.")

        embeddings = [embeds_obj.embedding for embeds_obj in embeds.data]
        return embeddings



from semantic_router.route import Route
from semantic_router.routers import HybridRouter, RouterConfig, SemanticRouter

__all__ = ["SemanticRouter", "HybridRouter", "Route", "RouterConfig"]

__version__ = "0.1.2"



import inspect
from typing import Any, Callable, ClassVar, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from semantic_router.llms import BaseLLM
from semantic_router.schema import Message, RouteChoice
from semantic_router.utils.logger import logger


class Parameter(BaseModel):
    """Parameter for a function.

    :param name: The name of the parameter.
    :type name: str
    :param description: The description of the parameter.
    :type description: Optional[str]
    :param type: The type of the parameter.
    :type type: str
    :param default: The default value of the parameter.
    :type default: Any
    :param required: Whether the parameter is required.
    :type required: bool
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)

    name: str = Field(description="The name of the parameter")
    description: Optional[str] = Field(
        default=None, description="The description of the parameter"
    )
    type: str = Field(description="The type of the parameter")
    default: Any = Field(description="The default value of the parameter")
    required: bool = Field(description="Whether the parameter is required")

    def to_ollama(self):
        """Convert the parameter to a dictionary for an Ollama-compatible function schema.

        :return: The parameter in dictionary format.
        :rtype: Dict[str, Any]
        """
        return {
            self.name: {
                "description": self.description,
                "type": self.type,
            }
        }


class FunctionSchema:
    """Class that consumes a function and can return a schema required by
    different LLMs for function calling.
    """

    name: str = Field(description="The name of the function")
    description: str = Field(description="The description of the function")
    signature: str = Field(description="The signature of the function")
    output: str = Field(description="The output of the function")
    parameters: List[Parameter] = Field(description="The parameters of the function")

    def __init__(self, function: Union[Callable, BaseModel]):
        """Initialize the FunctionSchema.

        :param function: The function to consume.
        :type function: Union[Callable, BaseModel]
        """
        self.function = function
        if callable(function):
            self._process_function(function)
        elif isinstance(function, BaseModel):
            raise NotImplementedError("Pydantic BaseModel not implemented yet.")
        else:
            raise TypeError("Function must be a Callable or BaseModel")

    def _process_function(self, function: Callable):
        """Process the function to get the name, description, signature, and output.

        :param function: The function to process.
        :type function: Callable
        """
        self.name = function.__name__
        self.description = str(inspect.getdoc(function))
        self.signature = str(inspect.signature(function))
        self.output = str(inspect.signature(function).return_annotation)
        parameters = []
        for param in inspect.signature(function).parameters.values():
            parameters.append(
                Parameter(
                    name=param.name,
                    type=param.annotation.__name__,
                    default=param.default,
                    required=False if param.default is param.empty else True,
                )
            )
        self.parameters = parameters

    def to_ollama(self):
        """Convert the FunctionSchema to an Ollama-compatible function schema dictionary.

        :return: The function schema in dictionary format.
        :rtype: Dict[str, Any]
        """
        schema_dict = {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": {
                    "type": "object",
                    "properties": {
                        param.name: {
                            "description": (
                                param.description
                                if isinstance(param.description, str)
                                else None
                            ),
                            "type": self._ollama_type_mapping(param.type),
                        }
                        for param in self.parameters
                    },
                    "required": [
                        param.name for param in self.parameters if param.required
                    ],
                },
            },
        }
        return schema_dict

    def _ollama_type_mapping(self, param_type: str) -> str:
        """Map the parameter type to an Ollama-compatible type.

        :param param_type: The type of the parameter.
        :type param_type: str
        :return: The Ollama-compatible type.
        :rtype: str
        """
        if param_type == "int":
            return "number"
        elif param_type == "float":
            return "number"
        elif param_type == "str":
            return "string"
        elif param_type == "bool":
            return "boolean"
        else:
            return "object"


def get_schema_list(items: List[Union[BaseModel, Callable]]) -> List[Dict[str, Any]]:
    """Get a list of function schemas from a list of functions or Pydantic BaseModels.

    :param items: The functions or BaseModels to get the schemas for.
    :type items: List[Union[BaseModel, Callable]]
    :return: A list of function schemas.
    :rtype: List[Dict[str, Any]]
    """
    schemas = []
    for item in items:
        schema = get_schema(item)
        schemas.append(schema)
    return schemas


def get_schema(item: Union[BaseModel, Callable]) -> Dict[str, Any]:
    """Get a function schema from a function or Pydantic BaseModel.

    :param item: The function or BaseModel to get the schema for.
    :type item: Union[BaseModel, Callable]
    :return: The function schema.
    :rtype: Dict[str, Any]
    """
    if isinstance(item, BaseModel):
        signature_parts = []
        for field_name, field_model in item.__annotations__.items():
            field_info = item.__fields__[field_name]
            default_value = field_info.default

            if default_value:
                default_repr = repr(default_value)
                signature_part = (
                    f"{field_name}: {field_model.__name__} = {default_repr}"
                )
            else:
                signature_part = f"{field_name}: {field_model.__name__}"

            signature_parts.append(signature_part)
        signature = f"({', '.join(signature_parts)}) -> str"
        schema = {
            "name": item.__class__.__name__,
            "description": item.__doc__,
            "signature": signature,
        }
    else:
        schema = {
            "name": item.__name__,
            "description": str(inspect.getdoc(item)),
            "signature": str(inspect.signature(item)),
            "output": str(inspect.signature(item).return_annotation),
        }
    return schema


def convert_python_type_to_json_type(param_type: str) -> str:
    """Convert a Python type to a JSON type.

    :param param_type: The type of the parameter.
    :type param_type: str
    :return: The JSON type.
    :rtype: str
    """
    if param_type == "int":
        return "number"
    if param_type == "float":
        return "number"
    if param_type == "str":
        return "string"
    if param_type == "bool":
        return "boolean"
    if param_type == "NoneType":
        return "null"
    if param_type == "list":
        return "array"
    else:
        return "object"


# TODO: Add route layer object to the input, solve circular import issue
async def route_and_execute(
    query: str, llm: BaseLLM, functions: List[Callable], layer
) -> Any:
    """Route and execute a function.

    :param query: The query to route and execute.
    :type query: str
    :param llm: The LLM to use.
    :type llm: BaseLLM
    :param functions: The functions to execute.
    :type functions: List[Callable]
    :param layer: The layer to use.
    :type layer: Layer
    :return: The result of the function.
    :rtype: Any
    """
    route_choice: RouteChoice = layer(query)

    for function in functions:
        if function.__name__ == route_choice.name:
            if route_choice.function_call:
                return function(**route_choice.function_call[0])

    logger.warning("No function found, calling LLM.")
    llm_input = [Message(role="user", content=query)]
    return llm(llm_input)



import os
from asyncio import sleep as asleep
from time import sleep
from typing import Any, List, Optional, Union

import openai
import tiktoken
from openai import OpenAIError
from openai._types import NotGiven
from openai.types import CreateEmbeddingResponse
from pydantic import PrivateAttr

from semantic_router.encoders import DenseEncoder
from semantic_router.schema import EncoderInfo
from semantic_router.utils.defaults import EncoderDefault
from semantic_router.utils.logger import logger

model_configs = {
    "text-embedding-ada-002": EncoderInfo(
        name="text-embedding-ada-002",
        token_limit=8192,
        threshold=0.82,
    ),
    "text-embedding-3-small": EncoderInfo(
        name="text-embedding-3-small",
        token_limit=8192,
        threshold=0.3,
    ),
    "text-embedding-3-large": EncoderInfo(
        name="text-embedding-3-large",
        token_limit=8192,
        threshold=0.3,
    ),
}


class OpenAIEncoder(DenseEncoder):
    """OpenAI encoder class for generating embeddings using OpenAI API.

    The OpenAIEncoder class is a subclass of DenseEncoder and utilizes the OpenAI API
    to generate embeddings for given documents. It requires an OpenAI API key and
    supports customization of the score threshold for filtering or processing the embeddings.
    """

    _client: Optional[openai.Client] = PrivateAttr(default=None)
    _async_client: Optional[openai.AsyncClient] = PrivateAttr(default=None)
    dimensions: Union[int, NotGiven] = NotGiven()
    token_limit: int = 8192  # default value, should be replaced by config
    _token_encoder: Any = PrivateAttr()
    type: str = "openai"
    max_retries: int = 3

    def __init__(
        self,
        name: Optional[str] = None,
        openai_base_url: Optional[str] = None,
        openai_api_key: Optional[str] = None,
        openai_org_id: Optional[str] = None,
        score_threshold: Optional[float] = None,
        dimensions: Union[int, NotGiven] = NotGiven(),
        max_retries: int = 3,
    ):
        """Initialize the OpenAIEncoder.

        :param name: The name of the embedding model to use.
        :type name: str
        :param openai_base_url: The base URL for the OpenAI API.
        :type openai_base_url: str
        :param openai_api_key: The OpenAI API key.
        :type openai_api_key: str
        :param openai_org_id: The OpenAI organization ID.
        :type openai_org_id: str
        :param score_threshold: The score threshold for the embeddings.
        :type score_threshold: float
        :param dimensions: The dimensions of the embeddings.
        :type dimensions: int
        :param max_retries: The maximum number of retries for the OpenAI API call.
        :type max_retries: int
        """
        if name is None:
            name = EncoderDefault.OPENAI.value["embedding_model"]
        if score_threshold is None and name in model_configs:
            set_score_threshold = model_configs[name].threshold
        elif score_threshold is None:
            logger.warning(
                f"Score threshold not set for model: {name}. Using default value."
            )
            set_score_threshold = 0.82
        else:
            set_score_threshold = score_threshold
        super().__init__(
            name=name,
            score_threshold=set_score_threshold,
        )
        api_key = openai_api_key or os.getenv("OPENAI_API_KEY")
        base_url = openai_base_url or os.getenv("OPENAI_BASE_URL")
        openai_org_id = openai_org_id or os.getenv("OPENAI_ORG_ID")
        if api_key is None or api_key.strip() == "":
            raise ValueError("OpenAI API key cannot be 'None' or empty.")
        if max_retries is not None:
            self.max_retries = max_retries
        try:
            self._client = openai.Client(
                base_url=base_url, api_key=api_key, organization=openai_org_id
            )
            self._async_client = openai.AsyncClient(
                base_url=base_url, api_key=api_key, organization=openai_org_id
            )
        except Exception as e:
            raise ValueError(
                f"OpenAI API client failed to initialize. Error: {e}"
            ) from e
        # set dimensions to support openai embed 3 dimensions param
        self.dimensions = dimensions
        # if model name is known, set token limit
        if name in model_configs:
            self.token_limit = model_configs[name].token_limit
        # get token encoder
        self._token_encoder = tiktoken.encoding_for_model(name)

    def __call__(self, docs: List[str], truncate: bool = True) -> List[List[float]]:
        """Encode a list of text documents into embeddings using OpenAI API.

        :param docs: List of text documents to encode.
        :param truncate: Whether to truncate the documents to token limit. If
            False and a document exceeds the token limit, an error will be
            raised.
        :return: List of embeddings for each document."""
        if self._client is None:
            raise ValueError("OpenAI client is not initialized.")
        embeds = None

        if truncate:
            # check if any document exceeds token limit and truncate if so
            docs = [self._truncate(doc) for doc in docs]

        # Exponential backoff
        for j in range(self.max_retries + 1):
            try:
                embeds = self._client.embeddings.create(
                    input=docs,
                    model=self.name,
                    dimensions=self.dimensions,
                )
                if embeds.data:
                    break
            except OpenAIError as e:
                logger.error("Exception occurred", exc_info=True)
                if self.max_retries != 0 and j < self.max_retries:
                    sleep(2**j)
                    logger.warning(
                        f"Retrying in {2**j} seconds due to OpenAIError: {e}"
                    )
                else:
                    raise

            except Exception as e:
                logger.error(f"OpenAI API call failed. Error: {e}")
                raise ValueError(f"OpenAI API call failed. Error: {str(e)}") from e

        if (
            not embeds
            or not isinstance(embeds, CreateEmbeddingResponse)
            or not embeds.data
        ):
            logger.info(f"Returned embeddings: {embeds}")
            raise ValueError("No embeddings returned.")

        embeddings = [embeds_obj.embedding for embeds_obj in embeds.data]
        return embeddings

    def _truncate(self, text: str) -> str:
        """Truncate a document to the token limit.

        :param text: The document to truncate.
        :type text: str
        :return: The truncated document.
        :rtype: str
        """
        # we use encode_ordinary as faster equivalent to encode(text, disallowed_special=())
        tokens = self._token_encoder.encode_ordinary(text)
        if len(tokens) > self.token_limit:
            logger.warning(
                f"Document exceeds token limit: {len(tokens)} > {self.token_limit}"
                "\nTruncating document..."
            )
            text = self._token_encoder.decode(tokens[: self.token_limit - 1])
            logger.info(f"Trunc length: {len(self._token_encoder.encode(text))}")
            return text
        return text

    async def acall(self, docs: List[str], truncate: bool = True) -> List[List[float]]:
        """Encode a list of text documents into embeddings using OpenAI API asynchronously.

        :param docs: List of text documents to encode.
        :param truncate: Whether to truncate the documents to token limit. If
            False and a document exceeds the token limit, an error will be
            raised.
        :return: List of embeddings for each document."""
        if self._async_client is None:
            raise ValueError("OpenAI async client is not initialized.")
        embeds = None

        if truncate:
            # check if any document exceeds token limit and truncate if so
            docs = [self._truncate(doc) for doc in docs]

        # Exponential backoff
        for j in range(self.max_retries + 1):
            try:
                embeds = await self._async_client.embeddings.create(
                    input=docs,
                    model=self.name,
                    dimensions=self.dimensions,
                )
                if embeds.data:
                    break
            except OpenAIError as e:
                logger.error("Exception occurred", exc_info=True)
                if self.max_retries != 0 and j < self.max_retries:
                    await asleep(2**j)
                    logger.warning(
                        f"Retrying in {2**j} seconds due to OpenAIError: {e}"
                    )
                else:
                    raise

            except Exception as e:
                logger.error(f"OpenAI API call failed. Error: {e}")
                raise ValueError(f"OpenAI API call failed. Error: {e}") from e

        if (
            not embeds
            or not isinstance(embeds, CreateEmbeddingResponse)
            or not embeds.data
        ):
            logger.info(f"Returned embeddings: {embeds}")
            raise ValueError("No embeddings returned.")

        embeddings = [embeds_obj.embedding for embeds_obj in embeds.data]
        return embeddings



"""
This module provides the HFEndpointEncoder class to embeddings models using Huggingface's endpoint.

The HFEndpointEncoder class is a subclass of DenseEncoder and utilizes a specified Huggingface
endpoint to generate embeddings for given documents. It requires the URL of the Huggingface
API endpoint and an API key for authentication. The class supports customization of the score
threshold for filtering or processing the embeddings.

Example usage:

    from semantic_router.encoders.hfendpointencoder import HFEndpointEncoder

    encoder = HFEndpointEncoder(
        huggingface_url="https://api-inference.huggingface.co/models/BAAI/bge-large-en-v1.5",
        huggingface_api_key="your-hugging-face-api-key"
    )
    embeddings = encoder(["document1", "document2"])

Classes:
    HFEndpointEncoder: A class for generating embeddings using a Huggingface endpoint.
"""

import os
import time
from typing import Any, Dict, List, Optional

import requests
from pydantic import PrivateAttr

from semantic_router.encoders import DenseEncoder
from semantic_router.utils.logger import logger

# TODO: this should support local models, and we should have another class for remote
# inference endpoint models


class HuggingFaceEncoder(DenseEncoder):
    """HuggingFace encoder class for local embedding models. Models can be trained and
    loaded from private repositories, or from the Huggingface Hub. The class supports
    customization of the score threshold for filtering or processing the embeddings.

    Example usage:

    ```python
    from semantic_router.encoders import HuggingFaceEncoder

    encoder = HuggingFaceEncoder(
        name="sentence-transformers/all-MiniLM-L6-v2",
        device="cuda"
    )
    embeddings = encoder(["document1", "document2"])
    ```
    """

    name: str = "sentence-transformers/all-MiniLM-L6-v2"
    type: str = "huggingface"
    tokenizer_kwargs: Dict = {}
    model_kwargs: Dict = {}
    device: Optional[str] = None
    _tokenizer: Any = PrivateAttr()
    _model: Any = PrivateAttr()
    _torch: Any = PrivateAttr()

    def __init__(self, **data):
        if data.get("score_threshold") is None:
            data["score_threshold"] = 0.5
        super().__init__(**data)
        self._tokenizer, self._model = self._initialize_hf_model()

    def _initialize_hf_model(self):
        try:
            from transformers import AutoModel, AutoTokenizer
        except (ImportError, RuntimeError):
            raise ImportError(
                "Please install transformers to use HuggingFaceEncoder. "
                "You can install it with: "
                "`pip install semantic-router[local]`"
            )

        try:
            import torch
        except RuntimeError:
            raise ImportError(
                "Please install Pytorch to use HuggingFaceEncoder. "
                "You can install it with: "
                "`pip install semantic-router[local]`"
            )

        self._torch = torch

        tokenizer = AutoTokenizer.from_pretrained(
            self.name,
            **self.tokenizer_kwargs,
        )

        model = AutoModel.from_pretrained(self.name, **self.model_kwargs)

        if self.device:
            model.to(self.device)

        else:
            device = "cuda" if self._torch.cuda.is_available() else "cpu"
            model.to(device)
            self.device = device

        return tokenizer, model

    def __call__(
        self,
        docs: List[str],
        batch_size: int = 32,
        normalize_embeddings: bool = True,
        pooling_strategy: str = "mean",
    ) -> List[List[float]]:
        """Encode a list of documents into embeddings using the local Hugging Face model.

        :param docs: A list of documents to encode.
        :type docs: List[str]
        :param batch_size: The batch size for encoding.
        """
        all_embeddings = []
        for i in range(0, len(docs), batch_size):
            batch_docs = docs[i : i + batch_size]

            encoded_input = self._tokenizer(
                batch_docs, padding=True, truncation=True, return_tensors="pt"
            ).to(self.device)

            with self._torch.no_grad():
                model_output = self._model(**encoded_input)

            if pooling_strategy == "mean":
                embeddings = self._mean_pooling(
                    model_output, encoded_input["attention_mask"]
                )
            elif pooling_strategy == "max":
                embeddings = self._max_pooling(
                    model_output, encoded_input["attention_mask"]
                )
            else:
                raise ValueError(
                    "Invalid pooling_strategy. Please use 'mean' or 'max'."
                )

            if normalize_embeddings:
                embeddings = self._torch.nn.functional.normalize(embeddings, p=2, dim=1)

            embeddings = embeddings.tolist()
            all_embeddings.extend(embeddings)
        return all_embeddings

    def _mean_pooling(self, model_output, attention_mask):
        """Perform mean pooling on the token embeddings.

        :param model_output: The output of the model.
        :type model_output: torch.Tensor
        :param attention_mask: The attention mask.
        """
        token_embeddings = model_output[0]
        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        )
        return self._torch.sum(
            token_embeddings * input_mask_expanded, 1
        ) / self._torch.clamp(input_mask_expanded.sum(1), min=1e-9)

    def _max_pooling(self, model_output, attention_mask):
        """Perform max pooling on the token embeddings.

        :param model_output: The output of the model.
        :type model_output: torch.Tensor
        :param attention_mask: The attention mask.
        """
        token_embeddings = model_output[0]
        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        )
        token_embeddings[input_mask_expanded == 0] = -1e9
        return self._torch.max(token_embeddings, 1)[0]


class HFEndpointEncoder(DenseEncoder):
    """HFEndpointEncoder class to embeddings models using Huggingface's inference endpoints.

    The HFEndpointEncoder class is a subclass of DenseEncoder and utilizes a specified
    Huggingface endpoint to generate embeddings for given documents. It requires the URL
    of the Huggingface API endpoint and an API key for authentication. The class supports
    customization of the score threshold for filtering or processing the embeddings.

    Example usage:

    ```python
    from semantic_router.encoders import HFEndpointEncoder

    encoder = HFEndpointEncoder(
        huggingface_url="https://api-inference.huggingface.co/models/BAAI/bge-large-en-v1.5",
        huggingface_api_key="your-hugging-face-api-key"
    )
    embeddings = encoder(["document1", "document2"])
    ```
    """

    name: str = "hugging_face_custom_endpoint"
    huggingface_url: Optional[str] = None
    huggingface_api_key: Optional[str] = None

    def __init__(
        self,
        name: Optional[str] = "hugging_face_custom_endpoint",
        huggingface_url: Optional[str] = None,
        huggingface_api_key: Optional[str] = None,
        score_threshold: float = 0.8,
    ):
        """Initializes the HFEndpointEncoder with the specified parameters.

        :param name: The name of the encoder.
        :type name: str
        :param huggingface_url: The URL of the Hugging Face API endpoint.
        :type huggingface_url: str
        :param huggingface_api_key: The API key for the Hugging Face API.
        :type huggingface_api_key: str
        :param score_threshold: A threshold for processing the embeddings.
        :type score_threshold: float
        :raise ValueError: If either `huggingface_url` or `huggingface_api_key` is None.
        """
        huggingface_url = huggingface_url or os.getenv("HF_API_URL")
        huggingface_api_key = huggingface_api_key or os.getenv("HF_API_KEY")
        if score_threshold is None:
            score_threshold = 0.8

        super().__init__(name=name, score_threshold=score_threshold)  # type: ignore

        if huggingface_url is None:
            raise ValueError("HuggingFace endpoint url cannot be 'None'.")
        if huggingface_api_key is None:
            raise ValueError("HuggingFace API key cannot be 'None'.")

        self.huggingface_url = huggingface_url or os.getenv("HF_API_URL")
        self.huggingface_api_key = huggingface_api_key or os.getenv("HF_API_KEY")

        try:
            self.query({"inputs": "Hello World!", "parameters": {}})
        except Exception as e:
            raise ValueError(
                f"HuggingFace endpoint client failed to initialize. Error: {e}"
            ) from e

    def __call__(self, docs: List[str]) -> List[List[float]]:
        """Encodes a list of documents into embeddings using the Hugging Face API.

        :param docs: A list of documents to encode.
        :type docs: List[str]
        :return: A list of embeddings for the given documents.
        :rtype: List[List[float]]
        :raise ValueError: If no embeddings are returned for a document.
        """
        embeddings = []
        for d in docs:
            try:
                output = self.query({"inputs": d, "parameters": {}})
                if not output or len(output) == 0:
                    raise ValueError("No embeddings returned from the query.")
                embeddings.append(output)

            except Exception as e:
                raise ValueError(
                    f"No embeddings returned for document. Error: {e}"
                ) from e
        return embeddings

    def query(self, payload, max_retries=3, retry_interval=5):
        """Sends a query to the Hugging Face API and returns the response.

        :param payload: The payload to send in the request.
        :type payload: dict
        :return: The response from the Hugging Face API.
        :rtype: dict
        :raise ValueError: If the query fails or the response status is not 200.
        """
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.huggingface_api_key}",
            "Content-Type": "application/json",
        }
        for attempt in range(1, max_retries + 1):
            try:
                response = requests.post(
                    self.huggingface_url,
                    headers=headers,
                    json=payload,
                    # timeout=timeout_seconds,
                )
                if response.status_code == 503:
                    estimated_time = response.json().get("estimated_time", "")
                    if estimated_time:
                        logger.info(
                            f"Model Initializing wait for - {estimated_time:.2f}s "
                        )
                        time.sleep(estimated_time)
                        continue
                else:
                    response.raise_for_status()

            except requests.exceptions.RequestException:
                if attempt < max_retries - 1:
                    logger.info(f"Retrying attempt: {attempt} for payload: {payload} ")
                    time.sleep(retry_interval)
                    retry_interval += attempt
                else:
                    raise ValueError(
                        f"Query failed with status {response.status_code}: {response.text}"
                    )

        return response.json()



import os
from typing import Any, List, Optional

from aurelio_sdk import AsyncAurelioClient, AurelioClient, EmbeddingResponse
from pydantic import Field

from semantic_router.encoders.base import AsymmetricSparseMixin, SparseEncoder
from semantic_router.schema import SparseEmbedding


class AurelioSparseEncoder(SparseEncoder, AsymmetricSparseMixin):
    """Sparse encoder using Aurelio Platform's embedding API. Requires an API key from
    https://platform.aurelio.ai
    """

    model: Optional[Any] = None
    client: AurelioClient = Field(default_factory=AurelioClient, exclude=True)
    async_client: AsyncAurelioClient = Field(
        default_factory=AsyncAurelioClient, exclude=True
    )
    type: str = "sparse"

    def __init__(
        self,
        name: str | None = None,
        api_key: Optional[str] = None,
    ):
        """Initialize the AurelioSparseEncoder.

        :param name: The name of the model to use.
        :type name: str | None
        :param api_key: The API key to use.
        :type api_key: str | None
        """
        if name is None:
            name = "bm25"
        super().__init__(name=name)
        if api_key is None:
            api_key = os.getenv("AURELIO_API_KEY")
        if api_key is None:
            raise ValueError("AURELIO_API_KEY environment variable is not set.")
        self.client = AurelioClient(api_key=api_key)
        self.async_client = AsyncAurelioClient(api_key=api_key)

    def __call__(self, docs: list[str]) -> list[SparseEmbedding]:
        """Encode a list of queries using the Aurelio Platform embedding API. Documents
        must be strings, sparse encoders do not support other types.
        """
        return self.encode_queries(docs)

    def encode_queries(self, docs: List[str]) -> List[SparseEmbedding]:
        res: EmbeddingResponse = self.client.embedding(
            input=docs, model=self.name, input_type="queries"
        )
        embeds = [SparseEmbedding.from_aurelio(r.embedding) for r in res.data]
        return embeds

    def encode_documents(self, docs: List[str]) -> List[SparseEmbedding]:
        res: EmbeddingResponse = self.client.embedding(
            input=docs, model=self.name, input_type="documents"
        )
        embeds = [SparseEmbedding.from_aurelio(r.embedding) for r in res.data]
        return embeds

    async def aencode_queries(self, docs: List[str]) -> list[SparseEmbedding]:
        res: EmbeddingResponse = await self.async_client.embedding(
            input=docs, model=self.name, input_type="queries"
        )
        embeds = [SparseEmbedding.from_aurelio(r.embedding) for r in res.data]
        return embeds

    async def aencode_documents(self, docs: List[str]) -> list[SparseEmbedding]:
        res: EmbeddingResponse = await self.async_client.embedding(
            input=docs, model=self.name, input_type="documents"
        )
        embeds = [SparseEmbedding.from_aurelio(r.embedding) for r in res.data]
        return embeds

    async def acall(self, docs: list[str]) -> list[SparseEmbedding]:
        """Asynchronously encode a list of documents using the Aurelio Platform
        embedding API. Documents must be strings, sparse encoders do not support other
        types.

        :param docs: The documents to encode.
        :type docs: list[str]
        :param input_type:
        :type semantic_router.encoders.encode_input_type.EncodeInputType
        :return: The encoded documents.
        :rtype: list[SparseEmbedding]
        """
        return await self.aencode_queries(docs)

    def fit(self, docs: List[str]):
        """Fit the encoder to a list of documents. AurelioSparseEncoder does not support
        fit yet.

        :param docs: The documents to fit the encoder to.
        :type docs: list[str]
        """
        raise NotImplementedError("AurelioSparseEncoder does not support fit.")



from typing import List, Optional

import requests

from semantic_router.llms import BaseLLM
from semantic_router.schema import Message
from semantic_router.utils.logger import logger


class OllamaLLM(BaseLLM):
    """LLM for Ollama. Enables fully local LLM use, helpful for local implementation of
    dynamic routes.
    """

    stream: bool = False

    def __init__(
        self,
        name: str = "openhermes",
        temperature: float = 0.2,
        max_tokens: Optional[int] = 200,
        stream: bool = False,
    ):
        """Initialize the OllamaLLM.

        :param name: The name of the Ollama model to use.
        :type name: str
        :param temperature: The temperature of the LLM.
        :type temperature: float
        :param max_tokens: The maximum number of tokens to generate.
        :type max_tokens: Optional[int]
        :param stream: Whether to stream the response.
        :type stream: bool
        """
        super().__init__(name=name)
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.stream = stream

    def __call__(
        self,
        messages: List[Message],
        temperature: Optional[float] = None,
        name: Optional[str] = None,
        max_tokens: Optional[int] = None,
        stream: Optional[bool] = None,
    ) -> str:
        """Call the OllamaLLM.

        :param messages: The messages to pass to the OllamaLLM.
        :type messages: List[Message]
        :param temperature: The temperature of the LLM.
        :type temperature: Optional[float]
        :param name: The name of the Ollama model to use.
        :type name: Optional[str]
        :param max_tokens: The maximum number of tokens to generate.
        :type max_tokens: Optional[int]
        :param stream: Whether to stream the response.
        :type stream: Optional[bool]
        """
        # Use instance defaults if not overridden
        temperature = temperature if temperature is not None else self.temperature
        name = name if name is not None else self.name
        max_tokens = max_tokens if max_tokens is not None else self.max_tokens
        stream = stream if stream is not None else self.stream

        try:
            payload = {
                "model": name,
                "messages": [m.to_openai() for m in messages],
                "options": {"temperature": temperature, "num_predict": max_tokens},
                "format": "json",
                "stream": stream,
            }
            response = requests.post("http://localhost:11434/api/chat", json=payload)
            output = response.json()["message"]["content"]

            return output
        except Exception as e:
            logger.error(f"LLM error: {e}")
            raise Exception(f"LLM error: {e}") from e



import logging

import colorlog


class CustomFormatter(colorlog.ColoredFormatter):
    """Custom formatter for the logger."""

    def __init__(self):
        super().__init__(
            "%(log_color)s%(asctime)s %(levelname)s %(name)s %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            log_colors={
                "DEBUG": "cyan",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "bold_red",
            },
            reset=True,
            style="%",
        )


def add_coloured_handler(logger):
    """Add a coloured handler to the logger."""
    formatter = CustomFormatter()
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    return logger


def setup_custom_logger(name):
    """Setup a custom logger."""
    logger = logging.getLogger(name)

    if not logger.hasHandlers():
        add_coloured_handler(logger)
        logger.setLevel(logging.INFO)
        logger.propagate = False

    return logger


logger: logging.Logger = setup_custom_logger(__name__)
